#ifndef _WASABI_APE_H
#define _WASABI_APE_H

// from Monkey's Audio SDK
#include <Source/Shared/All.h>
#include <Source/Shared/IO.h>
#include <Source/MACLib/APEDecompress.h>

#include "../bfc/std.h"
#include "../studio/services/svc_mediaconverter.h"
#include "../studio/services/servicei.h"
#include "../studio/corecb.h"
#include "../studio/wac.h"
//#include "../attribs/cfgitemi.h"
//#include "../attribs/attrint.h"
#include "tags.h"

class APE : public svc_mediaConverterI, private CIO {
private: // CIO stuff
    virtual int Open ( const wchar_t *pName ) { return 0; }

    virtual int Close() { return 0; }

    virtual int Read ( void *pBuffer, unsigned int nBytesToRead, unsigned int *pBytesRead )
    {
        int done = m_reader->read ( (char *)pBuffer, nBytesToRead );
        if ( pBytesRead ) *pBytesRead = done;
        return 0;
    }

    virtual int Write ( const void *pBuffer, unsigned int nBytesToWrite, unsigned int *pBytesWritten )
    {
        if ( pBytesWritten ) *pBytesWritten = 0;
        return 1;
    }

    virtual int Seek ( int nDistance, unsigned int nMoveMode )
    {
        if ( nMoveMode == FILE_CURRENT )
            nDistance += m_reader->getPos();
        else if ( nMoveMode == FILE_END )
            nDistance += m_reader->getLength();

        return m_reader->seek ( nDistance );
    }

    virtual int Create ( const wchar_t *pName ) { return 0; }
    virtual int Delete() { return 0; }
    virtual int SetEOF() { return 1; }

    // attributes
    virtual int GetPosition() { return (int)m_reader->getPos(); }
    virtual int GetSize() { return (int)m_reader->getLength(); }
    virtual int GetName ( wchar_t *pBuffer )
    {
	    if ( pBuffer ) *pBuffer = '\0';
	    return 1;
    }

    enum { DELTA_SAMPLES = 1024 };

public:
    APE ();
    virtual ~APE ();

    // service
    static const char *getServiceName() { return "Monkey's Audio Decoder"; }
    virtual int canConvertFrom ( svc_fileReader *reader, const char *name, const char *chunktype );

    virtual const char *getConverterTo() { return "PCM"; }

    virtual int getInfos ( MediaInfo *infos );
    virtual int processData ( MediaInfo *infos, ChunkList *chunk_list, bool *killswitch );
    virtual int getLatency ( void ) { return 0; }

    // callbacks
    virtual int corecb_onSeeked ( int newpos );

private:
    char *buffer;
    int buffer_size;
    IAPEDecompress *pAPEDecompress;
    svc_fileReader *m_reader;
    int needseek;

    int init ( MediaInfo *infos );
};

#endif // _WASABI_APE_H
