#include "tags.h"

/* V A R I A B L E S */
static const char *GenreList[] = {
    "Blues", "Classic Rock", "Country", "Dance", "Disco", "Funk",
    "Grunge", "Hip-Hop", "Jazz", "Metal", "New Age", "Oldies",
    "Other", "Pop", "R&B", "Rap", "Reggae", "Rock",
    "Techno", "Industrial", "Alternative", "Ska", "Death Metal", "Pranks",
    "Soundtrack", "Euro-Techno", "Ambient", "Trip-Hop", "Vocal", "Jazz+Funk",
    "Fusion", "Trance", "Classical", "Instrumental", "Acid", "House",
    "Game", "Sound Clip", "Gospel", "Noise", "AlternRock", "Bass",
    "Soul", "Punk", "Space", "Meditative", "Instrumental Pop", "Instrumental Rock",
    "Ethnic", "Gothic", "Darkwave", "Techno-Industrial", "Electronic", "Pop-Folk",
    "Eurodance", "Dream", "Southern Rock", "Comedy", "Cult", "Gangsta",
    "Top 40", "Christian Rap", "Pop/Funk", "Jungle", "Native American", "Cabaret",
    "New Wave", "Psychadelic", "Rave", "Showtunes", "Trailer", "Lo-Fi",
    "Tribal", "Acid Punk", "Acid Jazz", "Polka", "Retro", "Musical",
    "Rock & Roll", "Hard Rock", "Folk", "Folk/Rock", "National Folk", "Swing",
    "Fast-Fusion", "Bebob", "Latin", "Revival", "Celtic", "Bluegrass", "Avantgarde",
    "Gothic Rock", "Progressive Rock", "Psychedelic Rock", "Symphonic Rock", "Slow Rock", "Big Band",
    "Chorus", "Easy Listening", "Acoustic", "Humour", "Speech", "Chanson",
    "Opera", "Chamber Music", "Sonata", "Symphony", "Booty Bass", "Primus",
    "Porn Groove", "Satire", "Slow Jam", "Club", "Tango", "Samba",
    "Folklore", "Ballad", "Power Ballad", "Rhythmic Soul", "Freestyle", "Duet",
    "Punk Rock", "Drum Solo", "A capella", "Euro-House", "Dance Hall",
    "Goa", "Drum & Bass", "Club House", "Hardcore", "Terror",
    "Indie", "BritPop", "NegerPunk", "Polsk Punk", "Beat",
    "Christian Gangsta", "Heavy Metal", "Black Metal", "Crossover", "Contemporary C",
    "Christian Rock", "Merengue", "Salsa", "Thrash Metal", "Anime", "JPop",
    "SynthPop",
};

static void GenreToString ( char *GenreStr, const unsigned int genre )
{
    GenreStr[0] = '\0';
    if ( genre < sizeof(GenreList)/sizeof(*GenreList) ) STRCPY ( GenreStr, GenreList[genre] );
}

static void memcpy_crop ( char *dst, const char *src, size_t len )
{
    for ( size_t i = 0; i < len; i++ ) {
        if ( src[i] != '\0' ) dst[i] = src[i]; else break;
    }

    // dst[i] points behind the string contents
    while ( i > 0 && dst[i-1] == ' ' ) i--;

    dst[i] = '\0';
}

static int unicodeToUtf8 ( const WCHAR *lpWideCharStr, char *lpMultiByteStr, int cwcChars )
{
    const unsigned short *pwc = (unsigned short *)lpWideCharStr;
    unsigned char        *pmb = (unsigned char *)lpMultiByteStr;
    const unsigned short *pwce;
    size_t cBytes = 0;

    if ( cwcChars >= 0 ) {
        pwce = pwc + cwcChars;
    } else {
        pwce = (unsigned short *)((size_t)-1);
    }

    while ( pwc < pwce ) {
        unsigned short  wc = *pwc++;

        if ( wc < 0x00000080 ) {
            *pmb++ = (char)wc;
            cBytes++;
        } else
        if ( wc < 0x00000800 ) {
            *pmb++ = (char)(0xC0 | ((wc >>  6) & 0x1F));
            cBytes++;
            *pmb++ = (char)(0x80 |  (wc        & 0x3F));
            cBytes++;
        } else
        if ( wc < 0x00010000 ) {
            *pmb++ = (char)(0xE0 | ((wc >> 12) & 0x0F));
            cBytes++;
            *pmb++ = (char)(0x80 | ((wc >>  6) & 0x3F));
            cBytes++;
            *pmb++ = (char)(0x80 |  (wc        & 0x3F));
            cBytes++;
        }
        if ( wc == L'\0' )
            return cBytes;
    }

    return cBytes;
}

static int ConvertANSIToUTF8 ( const char *ansi, char *utf8 )
{
    WCHAR  *wszValue;
    size_t ansi_len;
    size_t len;

    *utf8 = '\0';
    if ( !ansi ) return 0;

    ansi_len = strlen ( ansi );

    if ( (wszValue = (WCHAR *)malloc ( (ansi_len + 1) * 2 )) == NULL ) return 0;

    // Convert ANSI value to Unicode
    if ( (len = MultiByteToWideChar ( CP_ACP, 0, ansi, ansi_len + 1, wszValue, (ansi_len + 1) * 2 )) == 0 ) {
        free ( wszValue );
        return 0;
    }

    // Convert Unicode value to UTF-8
    if ( (len = unicodeToUtf8 ( wszValue, utf8, -1 )) == 0 ) {
        free ( wszValue );
        return 0;
    }

    free ( wszValue );

    return 1;
}

struct APETagFooterStruct {
    unsigned char   ID       [8];    // should equal 'APETAGEX'
    unsigned char   Version  [4];    // 1000 = version 1.0, 2000 = version 2.0
    unsigned char   Length   [4];    // complete size of the tag, including footer, excluding header
    unsigned char   TagCount [4];    // number of fields in the tag
    unsigned char   Flags    [4];    // tag flags (none currently defined)
    unsigned char   Reserved [8];    // reserved for later use
};

static unsigned long Read_LE_Uint32 ( const unsigned char* p )
{
    return ((unsigned long)p[0] <<  0) |
           ((unsigned long)p[1] <<  8) |
           ((unsigned long)p[2] << 16) |
           ((unsigned long)p[3] << 24);
}

int ReadAPETag ( svc_fileReader *r, MediaInfo *infos, int &TagOffset, bool only_check )
{
    unsigned long               vsize;
    unsigned long               isize;
    unsigned long               flags;
    char                        *buff;
    char                        *p;
    char                        *end;
    struct APETagFooterStruct   T;
    unsigned long               TagLen;
    unsigned long               TagCount;
    int                         ape_ver;
    char                        *temp;
    int                         tag_fields = 0;

    r->seek ( TagOffset - sizeof T );
    if ( r->read ( (char *)&T, sizeof T ) != sizeof T ) return 0;
    if ( MEMCMP ( T.ID, "APETAGEX", sizeof T.ID ) ) return 0;
    ape_ver = Read_LE_Uint32 ( T.Version );
    if ( ape_ver != 2000 && ape_ver != 1000 ) return 0;
    TagLen = Read_LE_Uint32 ( T.Length );
    if ( TagLen <= sizeof T ) return 0;

    if ( only_check ) return 1;

    r->seek ( TagOffset - TagLen );
    if ( (buff = (char *)MALLOC (TagLen)) == NULL ) return 0;
    if ( (int)r->read (buff, TagLen - sizeof T) != (int)(TagLen - sizeof T) ) {
        free ( buff );
        return 0;
    }

    TagCount = Read_LE_Uint32 ( T.TagCount );
    end = buff + TagLen - sizeof T;
    for ( p = buff; p < end && TagCount--; ) {
        vsize = Read_LE_Uint32 ( (unsigned char*)p ); p += 4;
        flags = Read_LE_Uint32 ( (unsigned char*)p ); p += 4;
        isize = STRLEN ( p );

        if ( vsize > 0 ) {
            if ( ape_ver == 2000 ) {
                if ( !(flags & 2) ) { // not binary
                    temp = (char *)MALLOC ( vsize + 1 );
                    if ( temp ) {
                        memcpy ( temp, p + isize + 1, vsize );
                        temp[vsize] = '\0';

                        int fields = 0;

                        {
                            const char *t = temp;
                            const char *te = t + vsize;

                            while ( t < te ) {
                                if ( *t++ == '\0' ) fields++;
                            }
                        }

                        infos->setData ( p, temp, MediaInfo::INFO_VISIBLE );
                        tag_fields++;

                        if ( fields > 0 ) {
                            const char *t = temp;

                            while ( fields-- > 0 ) {
                                t = STRCHR ( t, '\0' );
                                if ( t ) {
                                    t++;
                                    infos->setData ( p, t, MediaInfo::INFO_VISIBLE );
                                    tag_fields++;
                                }
                            }
                        }

                        FREE ( temp );
                    }
                }
            } else {
                temp = (char *)MALLOC ( (vsize + 1) * 3 );
                if ( temp ) {
                    ConvertANSIToUTF8 ( p + isize + 1, temp );
                    infos->setData ( p, temp, MediaInfo::INFO_VISIBLE );
                    tag_fields++;
                    FREE ( temp );
                }
            }
        }

        p += isize + 1 + vsize;
    }

    FREE ( buff );

    TagOffset -= TagLen;
    if ( (ape_ver == 2000) && (Read_LE_Uint32 (T.Flags) & (1<<31)) ) TagOffset -= sizeof T; // Tag contains header

    return tag_fields;
}

int ReadID3v1Tag ( svc_fileReader *r, MediaInfo *infos, int &TagOffset, bool only_check )
{
    char tmp[128];
    char value[32];
    char utf8[32*3];
    int tag_fields = 0;

    r->seek ( TagOffset - sizeof (tmp) );
    if ( r->read  (tmp, sizeof (tmp)) != sizeof (tmp) ) return 0;
    if ( MEMCMP (tmp, "TAG", 3) ) return 0;

    if ( only_check ) return 1;

    memcpy_crop ( value, tmp +  3, 30 );
    if ( value[0] != '\0' ) {
        ConvertANSIToUTF8 ( value, utf8 );
        infos->setData ( APE_TAG_FIELD_TITLE, utf8, MediaInfo::INFO_VISIBLE );
        tag_fields++;
    }
    memcpy_crop ( value, tmp + 33, 30 );
    if ( value[0] != '\0' ) {
        ConvertANSIToUTF8 ( value, utf8 );
        infos->setData ( APE_TAG_FIELD_ARTIST, utf8, MediaInfo::INFO_VISIBLE );
        tag_fields++;
    }
    memcpy_crop ( value, tmp + 63, 30 );
    if ( value[0] != '\0' ) {
        ConvertANSIToUTF8 ( value, utf8 );
        infos->setData ( APE_TAG_FIELD_ALBUM, utf8, MediaInfo::INFO_VISIBLE );
        tag_fields++;
    }
    memcpy_crop ( value, tmp + 93,  4 );
    if ( value[0] != '\0' ) {
        ConvertANSIToUTF8 ( value, utf8 );
        infos->setData ( APE_TAG_FIELD_YEAR, utf8, MediaInfo::INFO_VISIBLE );
        tag_fields++;
    }
    memcpy_crop ( value, tmp + 97, 30 );
    if ( value[0] != '\0' ) {
        ConvertANSIToUTF8 ( value, utf8 );
        infos->setData ( APE_TAG_FIELD_COMMENT, utf8, MediaInfo::INFO_VISIBLE );
        tag_fields++;
    }

    if ( tmp[125] == 0 ) {
        SPRINTF ( value, "%i", (int)tmp[126] );
        if ( value[0] != '\0' && atoi (value) != 0 ) {
            infos->setData ( APE_TAG_FIELD_TRACK, value, MediaInfo::INFO_VISIBLE );
            tag_fields++;
        }
    }
    GenreToString ( value, tmp[127] );
    if ( value[0] != '\0' ) {
        ConvertANSIToUTF8 ( value, utf8 );
        infos->setData ( APE_TAG_FIELD_GENRE, utf8, MediaInfo::INFO_VISIBLE );
        tag_fields++;
    }

    TagOffset -= 128;

    return tag_fields;
}

int ReadTags ( MediaInfo *infos )
{
    svc_fileReader *r = infos->getReader();
    if ( !r || !r->canSeek() ) return 0;

    int orig_pos = r->getPos();

    int OldOffset;
    int TagOffset = r->getLength();
    int best_tag = TagOffset;

    // search for best tag type
    do {
        OldOffset = TagOffset;
        if ( ReadAPETag (r, infos, TagOffset, true) ) {
            best_tag = TagOffset;
            break;
        }
        ReadID3v1Tag ( r, infos, TagOffset, true );
    } while ( OldOffset != TagOffset );

    int tag_fields = 0;
    tag_fields += ReadID3v1Tag ( r, infos, best_tag, false );
    tag_fields += ReadAPETag ( r, infos, best_tag, false );

    r->seek ( orig_pos );

    return tag_fields;
}

int SkipID3v2Tag ( svc_fileReader *r )
{
    unsigned char tmp[10];

    if ( r->read ((char *)tmp, sizeof (tmp)) != sizeof (tmp) ) {
        r->seek ( 0 );
        return -1;
    }

    // check id3-tag
    if ( memcmp (tmp, "ID3", 3) ) {
        r->seek ( 0 );
        return 0;
    }

    // read flags
    int Unsynchronisation = tmp[5] & 0x80;
    int ExtHeaderPresent  = tmp[5] & 0x40;
    int ExperimentalFlag  = tmp[5] & 0x20;
    int FooterPresent     = tmp[5] & 0x10;

    if ( tmp[5] & 0x0F ) {                              // not (yet???) allowed
        r->seek ( 0 );
        return 0;
    }

    if ( (tmp[6] | tmp[7] | tmp[8] | tmp[9]) & 0x80 ) { // not allowed
        r->seek ( 0 );
        return 0;
    }

    // read HeaderSize (syncsave: 4 * $0xxxxxxx = 28 significant bits)
    unsigned int ret;
    ret  = tmp[6] << 21;
    ret += tmp[7] << 14;
    ret += tmp[8] <<  7;
    ret += tmp[9];
    ret += 10;
    if ( FooterPresent ) ret += 10;

    r->seek ( ret );

    return ret;
}
